@@ -979,6 +979,34 @@ public class TestNonBlockingAPI extends TomcatBaseTest {
         }
     }
 
+    @Test
+    public void testDelayedNBReadWrite() throws Exception {
+        Tomcat tomcat = getTomcatInstance();
+
+        Context ctx = tomcat.addContext("", null);
+        CountDownLatch latch1 = new CountDownLatch(2);
+        DelayedNBReadWriteServlet servlet = new DelayedNBReadWriteServlet(latch1);
+        String servletName = DelayedNBReadWriteServlet.class.getName();
+        Tomcat.addServlet(ctx, servletName, servlet);
+        ctx.addServletMappingDecoded("/", servletName);
+
+        tomcat.start();
+
+        CountDownLatch latch2 = new CountDownLatch(1);
+        List<Throwable> exceptions = new ArrayList<>();
+
+        Thread t = new Thread(
+            new RequestPostExecutor("http://localhost:" + getPort() + "/", latch2, exceptions));
+        t.start();
+
+        latch1.await(3000, TimeUnit.MILLISECONDS);
+        latch2.await(3000, TimeUnit.MILLISECONDS);
+
+        if (exceptions.size() > 0) {
+            Assert.fail();
+        }
+    }
+
     private static final class RequestExecutor implements Runnable {
         private final String url;
         private final CountDownLatch latch;
@@ -1007,6 +1035,34 @@ public class TestNonBlockingAPI extends TomcatBaseTest {
 
     }
 
+    private static final class RequestPostExecutor implements Runnable {
+        private final String url;
+        private final CountDownLatch latch;
+        private final List<Throwable> exceptions;
+
+        public RequestPostExecutor(String url, CountDownLatch latch, List<Throwable> exceptions) {
+            this.url = url;
+            this.latch = latch;
+            this.exceptions = exceptions;
+        }
+
+        @Override
+        public void run() {
+            try {
+                ByteChunk result = new ByteChunk();
+                int rc = postUrl("body".getBytes("utf-8"), url, result, null);
+                Assert.assertEquals(HttpServletResponse.SC_OK, rc);
+                Assert.assertTrue(result.toString().contains("OK"));
+            } catch (Throwable e) {
+                e.printStackTrace();
+                exceptions.add(e);
+            } finally {
+                latch.countDown();
+            }
+        }
+
+    }
+
     @WebServlet(asyncSupported = true)
     private static final class DelayedNBWriteServlet extends TesterServlet {
         private static final long serialVersionUID = 1L;
@@ -1038,6 +1094,105 @@ public class TestNonBlockingAPI extends TomcatBaseTest {
 
     }
 
+    @WebServlet(asyncSupported = true)
+    private static final class DelayedNBReadWriteServlet extends TesterServlet {
+        private static final long serialVersionUID = 1L;
+        private final transient CountDownLatch latch;
+
+        public DelayedNBReadWriteServlet(CountDownLatch latch) {
+            this.latch = latch;
+        }
+
+        @Override
+        protected void doPost(HttpServletRequest request, HttpServletResponse response)
+            throws ServletException, IOException {
+            final AsyncContext ctx = request.startAsync();
+            ctx.setTimeout(1000);
+
+            Thread readWriteListener =  new Thread(new ReadWriteListener(latch, ctx));
+            readWriteListener.start();
+        }
+    }
+
+    private static final class ReadWriteListener implements Runnable {
+        private final transient CountDownLatch latch;
+        private final transient AsyncContext ctx;
+
+        public ReadWriteListener(CountDownLatch latch, AsyncContext ctx){
+            this.latch = latch;
+            this.ctx = ctx;
+        }
+
+        @Override
+        public void run() {
+            try {
+                setListeners();
+            } catch (IOException e) {
+                e.printStackTrace();
+            }
+        }
+
+        private void setListeners() throws IOException {
+            final ServletInputStream is = ctx.getRequest().getInputStream();
+            final ServletOutputStream os = ctx.getResponse().getOutputStream();
+
+            is.setReadListener(new ReadListener() {
+                @Override
+                public void onDataAvailable() {
+
+                    try {
+                        byte buffer[] = new byte[1 * 4];
+                        while (is.isReady() && !is.isFinished()) {
+                            is.read(buffer);
+                        }
+                        String body = new String(buffer, StandardCharsets.UTF_8);
+                        Assert.assertTrue(body.equals("body"));
+
+                    } catch (IOException ex) {
+                        ex.printStackTrace();
+                    }
+                }
+
+                @Override
+                public void onAllDataRead() {
+                    latch.countDown();
+                }
+
+                @Override
+                public void onError(Throwable t) {
+                }
+            });
+
+            os.setWriteListener(new WriteListener() {
+                private boolean written = false;
+
+                @Override
+                public void onWritePossible() throws IOException {
+                    ServletOutputStream out = ctx.getResponse().getOutputStream();
+                    if (out.isReady() && !written) {
+                        out.println("OK");
+                        written = true;
+                    }
+                    if (out.isReady() && written) {
+                        out.flush();
+                        if (out.isReady()) {
+                            ctx.complete();
+                            latch.countDown();
+                        }
+                    }
+                }
+
+                @Override
+                public void onError(Throwable t) {
+                    t.printStackTrace();
+                }
+
+            });
+        }
+
+    }
+
+
     private static final class Emitter implements Serializable {
 
         private static final long serialVersionUID = 1L;
@@ -1539,4 +1694,4 @@ public class TestNonBlockingAPI extends TomcatBaseTest {
             // NO-OP
         }
     }
-}
+}
\ No newline at end of file
